from google.cloud import pubsub
import hashlib
import time
import json
from datetime import datetime
import pytz


class Pubsub(object):

    def __init__(self, project_id=None, publish_batch_max_messages=100, publish_batch_max_baytes=1024, publish_batch_max_latency=10):
        self.project_id = project_id
        self.batch_pubsub_settings = pubsub.types.BatchSettings(
            max_messages=publish_batch_max_messages,
            max_bytes=publish_batch_max_baytes,  # in kB
            max_latency=publish_batch_max_latency,  # in ms
        )
        self.publisher_client = pubsub.PublisherClient(self.batch_pubsub_settings)
        self.futures = dict()

    @staticmethod
    def get_time(time_zone='UTC'):
        tz = pytz.timezone(time_zone)
        return datetime.now(tz)

    def create_topic(self, topic_name: str, project_id=None):
        if project_id is None:
            project_id = self.project_id
        topic_path = "projects/{}/topics/{}".format(project_id, topic_name)
        return self.publisher_client.create_topic(request={"name": topic_path})

    def create_pull_subscription(self, subscription_name: str, topic_path: str, project_id=None, ack_deadline_seconds=10):
        if project_id is None:
            project_id = self.project_id
        subscription_path = "projects/{}/subscriptions/{}".format(project_id, subscription_name)
        subscriber_client = pubsub.SubscriberClient()
        return subscriber_client.create_subscription(request={"name": subscription_path,
                                                              "topic": topic_path,
                                                              "ack_deadline_seconds": ack_deadline_seconds,
                                                              "expiration_policy": {"ttl": None}})

    def publish_messages(self, data, topic_path: str, wait_until_all_published=True, encoding="utf-8"):

        def get_callback(future, key):
            def callback(future):
                message_id = future.result()
                self.futures.pop(key)
            return callback

        for item in data:
            key = hashlib.sha256("{}-{}".format(item, str(self.get_time())).encode()).hexdigest()
            self.futures.update({key: None})
            # When you publish a message, the client returns a future.
            future = self.publisher_client.publish(topic_path, json.dumps(item).encode(encoding))
            self.futures[key] = future
            # Publish failures shall be handled in the callback function.
            future.add_done_callback((get_callback(future, key)))

        if wait_until_all_published:
            # Wait for all the publish futures to resolve before exiting.
            self.wait_for_publishing()

    def wait_for_publishing(self):
        while self.futures:
            time.sleep(0.01)

    def publish_message(self, data, topic_path: str, encoding="utf-8"):
        data = json.dumps(data).encode(encoding)
        future = self.publisher_client.publish(topic_path, data)
        return "{}".format(future.result())  # message_id


if __name__ == '__main__':


    # Init pubsub_handler with default project:
    pubsub_handler = Pubsub(project_id="playground-josef")


    # Create Pubsub Topic in default project:
    topic = pubsub_handler.create_topic("test-topic")
    print(topic)

    # Create Pubsub Topic in other project:
    topic = pubsub_handler.create_topic("test-topic", "spielwiese-tobias")
    print(topic)

    # Create pull Subscription in default project:
    subscription = pubsub_handler.create_pull_subscription(subscription_name="test-subs",
                                                           topic_path="projects/playground-josef/topics/temp-test",
                                                           ack_deadline_seconds=100)
    print(subscription)

    # Create pull Subscription in other project:
    subscription = pubsub_handler.create_pull_subscription(subscription_name="test-subs",
                                                           project_id="spielwiese-tobias",
                                                           topic_path="projects/spielwiese-tobias/topics/test-topic",
                                                           ack_deadline_seconds=30)
    print(subscription)


    # Publish messages - batch (if one bundle to publish) - optional: encoding (default="utf-8"):
    data = [{"column1": i, "column2": "value {}".format(i)} for i in range(10000)]
    pubsub_handler.publish_messages(data=data,
                                    topic_path="projects/playground-josef/topics/temp-test",
                                    wait_until_all_published=True)

    # Publish messages - batch (if more than one bundle to publish) - optional: encoding (default="utf-8"):
    data = [{"column1": i, "column2": "value {}".format(i)} for i in range(10000)]
    for i in range(10):
        pubsub_handler.publish_messages(data=data,
                                        topic_path="projects/playground-josef/topics/temp-test",
                                        wait_until_all_published=False)
    pubsub_handler.wait_for_publishing()
    
    # Publish message - just one message, waiting until message is published - optional: encoding (default="utf-8"). 
    data = {"column1": 1, "column2": "value 1"}
    message_id = pubsub_handler.publish_message(data=data, topic_path="projects/playground-josef/topics/temp-test")
    print(message_id)




from google.cloud import secretmanager


class SecretManager(object):

    def __init__(self, project_id: str, service_accout_json_file_path=None):
        """
        Init your google cloud secret manager handler
        :param project_id: str, required
        :param service_accout_json_file_path: str, optional. The file path to your service account json key file. If
                not specified, the client uses the default credentials from your environment.
        """

        self.project_id = project_id
        self.service_accout_json_file_path = service_accout_json_file_path

        # Create the Secret Manager client.
        if service_accout_json_file_path is None:
            self.client = secretmanager.SecretManagerServiceClient()
        else:
            self.client = secretmanager.SecretManagerServiceClient.from_service_account_file(service_accout_json_file_path)

        # Build the parent name from the project.
        self.project_path = "projects/{}".format(project_id)

    def create_secret(self, secret_id: str, value=None, regions=None, labels=None):
        """
        If you not specify a value, you only create the secret, without value.

        :param secret_id: str, required. The secret_id you want to create.
        :param value: str, optional. Your secret value.
        :param regions: list, optional. Your list of regions where you want replicas of your secret.
        :param labels: dict, optional. Your GCP labels.
        :return: secret_id name or secret_id version (str)

        """
        # Create the secret.
        replicas = [{'location': 'europe-west4'}]
        if isinstance(regions, list):
            replicas = [{'location': region} for region in regions]

        response = self.client.create_secret(
            request={
            'parent': self.project_path,
            'secret_id': secret_id,
            'secret': {'replication': {
                'user_managed':
                    {'replicas': replicas},
            },
            'labels': labels},

        })

        if value is None:
            return response.name
        else:
            # Add secret version
            return self.add_secret_version(secret_id, value)

    def add_secret_version(self, secret_id: str, value: str):
        """
        Add a secret version with a value to your existing secret

        :param secret_id: str, required.
        :param value: str, required. Your secret value.
        :return:  secret_id version (str)
        """
        # Convert the string payload into a bytes.
        payload = value.encode('UTF-8')
        parent = self.client.secret_path(self.project_id, secret_id)

        # Add the secret version.
        response = self.client.add_secret_version(request={"parent": parent, "payload": {'data': payload}})
        return response.name

    def access_secret_version(self, secret_id, version_id):
        """
        Get the value of your secret.

        :param secret_id: str, required
        :param version_id: str or int, required
        :return: secret data (str)  !!!WARNING: Do not print the secret in a production environment!!!
        """
        # Build the resource name of the secret version.
        name = "projects/{}/secrets/{}/versions/{}".format(self.project_id, secret_id, version_id)
        # Access the secret version.
        response = self.client.access_secret_version(request={"name": name})
        return response.payload.data.decode('UTF-8')

    def list_secret_versions(self, secret_id: str):
        """
        List all version of a specified secret.

        :param secret_id: str, required
        :return: list of all versions of your secret
        """
        name = "projects/{}/secrets/{}".format(self.project_id, secret_id)
        return [version for version in self.client.list_secret_versions(request={"parent": name})]

    def list_secrets(self):
        """
        List all secrets in the given project.

        :return: list of all secret names
        """
        # Build the resource name of the parent project.
        return [secret for secret in self.client.list_secrets({"parent": self.project_path})]

    def get_secret(self, secret_id):
        """
        Get information about the given secret. This only returns metadata about
        the secret container, not any secret material.

        :param secret_id: str, required
        :return: A :class:`~google.cloud.secretmanager_v1.types.Secret` instance.

        """
        # Build the resource name of the secret.
        name = self.client.secret_path(self.project_id, secret_id)

        # Get the secret.
        return self.client.get_secret(request={"name": name})

    def delete_secret(self, secret_id):
        """
        Delete the secret with the given name and all of its versions.

        :param secret_id: str, required
        """
        # Build the resource name of the secret.
        name = self.client.secret_path(self.project_id, secret_id)

        # Delete the secret.
        self.client.delete_secret(request={"name": name})


    def delete_secret_version(self, secret_id, secret_version):

        '''

        :param secret_id: str, required
        :param secret_version: str, required
        :return:
        '''

        name = "projects/{}/secrets/{}/versions/{}".format(self.project_id, secret_id, secret_version)
        self.client.destroy_secret_version(request={"name": name})




if __name__ == '__main__':


    # First, enable the secret manager API  on GCP console

    # Init secretmanager object with your the default credential from your environment:
    secretmanager = SecretManager("playground-josef")
    # Init secretmanager object with your a service account json file - specify the path to your file:
    #secretmanager = SecretManager("v135-5683-playground-goppold", "test.json")
    #secretmanager.update_secret_version('test', 'new value')
    
    # Create a secret 
    secret_name = secretmanager.create_secret("my-secret")
    print(secret_name)
    
    
    # Create a secret with value
    secret_version_name = secretmanager.create_secret("my-secret-with-value", "This is my secret value")
    print(secret_version_name)
    
    # By default: regions=["europe-west4"] and labels=None. Please specify this params of you need to.

    # Add a secret value to a existing secret:
    secret_version_name_1 = secretmanager.add_secret_version("my-secret", "This is another secret value")
    print(secret_version_name_1)
    
    secret_version_name_2 = secretmanager.add_secret_version("my-secret", "This is a new verison of my secret value")
    print(secret_version_name_2)

    # Accessing value of a secret version: (!!!WARNING: Do not print the secret in a production environment!!!)
    value_1 = secretmanager.access_secret_version("my-secret", 1)
    print(value_1)
    value_2 = secretmanager.access_secret_version("my-secret", 2)
    print(value_2)
    value_3 = secretmanager.access_secret_version("my-secret-with-value", 1)
    print(value_3)

    # List all version of a specified secret:
    secret_verions = secretmanager.list_secret_versions("my-secret")
    print(secret_verions)

    # List all secrets in my project:
    secrets = secretmanager.list_secrets()
    print(secrets)

    # Get metadata of a specific secret:
    metadata = secretmanager.get_secret("my-secret")
    print(metadata.name)

    # Delete a secret with the given name and all of its versions:
    secretmanager.delete_secret("my-secret")
    secretmanager.delete_secret("my-secret-with-value")




from kfp.v2 import components

from google.cloud import storage


class KubeFlowHandler(object):

    def __init__(self):
        self.storage_client = storage.Client()

    def load_kfp_component(self, filename: str, bucket_id: str = 'tadd-kubeflow-components') -> components:

        bucket = self.storage_client.get_bucket(bucket_id)

        blob = bucket.blob(filename)
        blob = blob.download_as_string()
        blob = blob.decode('utf-8')

        blob_string = str(blob)

        return components.load_component_from_text("""{}""".format(blob_string))


if __name__ == '__main__':


    kf = KubeFlowHandler()


    component = kf.load_kfp_component('tadd-kubeflow-components', 'test/component.yaml')


    print("test")














import requests
import json
import jwt
from jwt.algorithms import RSAAlgorithm
import time


class Fifa(object):

    def __init__(self,
                 jwk_url="https://auth.mediamarktsaturn.com/.well-known/jwks.json",
                 fifa_token_url="https://auth.mediamarktsaturn.com/oauth/token",
                 max_num_retries=5):

        self.jwk_url = jwk_url
        self.fifa_token_url = fifa_token_url
        self.max_num_retries = max_num_retries
        self.__keys = self.__get_keys()

    def __get_keys_from_api(self):
        try:
            r = requests.get(self.jwk_url)
            if r.status_code != 200:
                raise Exception("status_code={}, content={}".format(r.status_code, r.content.decode('utf-8')))

            jwks = json.loads(r.content.decode('utf-8'))
            public_keys = {}
            for jwk in jwks['keys']:
                kid = jwk['kid']
                public_keys[kid] = RSAAlgorithm.from_jwk(json.dumps(jwk))  # TODO maybe consider other algorithms
            return None, public_keys
        except Exception as exc:
            return "Error while getting jwk from {}: {}".format(self.jwk_url, exc), None

    def __get_keys(self, sleep_start=0.1):
        # Retry block, if i.e. the connection is broken
        for _ in range(self.max_num_retries):
            error, keys = self.__get_keys_from_api()
            if error:
                time.sleep(sleep_start)
                sleep_start = sleep_start*2
                continue  # try again
            else:
                break
        else:
            raise Exception("Error while getting jwk from {}. Max number of retries reached.".format(self.jwk_url))
        return keys

    def verify_jwt_token(self, jwt_token: str, audience):
        """
        Verify if a given jwt_token is valid.

        :param jwt_token: str, jwt token which has to validated
        :param audience: str or tuple of strings with the allowed audience
        :return: tuple(error, payload) -> if no error occurs validation was successful: error=None, payload=dict(...)
                                       -> if error occurs: validation failed: error=str(...), payload=None
        """
        try:
            kid = jwt.get_unverified_header(jwt_token)['kid']
            key = self.__keys.get(kid)
            if key is None:
                # Get potential new keys:
                self.__keys = self.__get_keys()
                key = self.__keys.get(kid)
                if key is None:
                    raise Exception("No public key found for kid: {}.".format(kid))
            payload = jwt.decode(jwt_token, key=key, algorithms=['RS256'], audience=audience)
            return None, payload
        except Exception as exc:
            return "Error while validating jwt: {}".format(exc), None

    def get_jwt_token(self, client_id: str, client_secret: str):
        """
        :param client_id: str
        :param client_secret: str
        :return: tuple (error, bearer_token). If no error: (None, "eyJh334...."). If error: ("Error message", None)
        """
        try:
            data = {'grant_type': 'client_credentials'}
            try:
                access_token_response = requests.post(self.fifa_token_url, data=data, auth=(client_id, client_secret))
            except Exception as ex:
                return 'Get new FIFA bearer token failed: {}'.format(ex), None

            status_code = access_token_response.status_code

            # Check response code
            if status_code != 200:
                return 'FIFA API call not successful: code {}, response {}'.format(status_code, access_token_response.text), None

            result_body = json.loads(access_token_response.content)
            return None,  str(result_body["access_token"])

        except Exception as exc:
            return "Unexpected error: {}".format(exc), None


if __name__ == '__main__':

    # Init fifa handler:
    fifa_handler = Fifa()
    # Optional paramaters (with default values):
    #   - jwk_url="https://auth.mediamarktsaturn.com/.well-known/jwks.json",
    #   - fifa_token_url="https://auth.mediamarktsaturn.com/oauth/token",
    #   - max_num_retries=5

    # Get fifa jwt token: (If error is None -> everything is ok)
    error, token = fifa_handler.get_jwt_token(client_id="...", client_secret="...")
    print(error, token)
    # Verify fifa jwt token: (If error is None -> everything is ok)
    error, payload = fifa_handler.verify_jwt_token(jwt_token=token, audience="...")
    print(error, payload)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

required_packages=[
    'google-cloud-logging>=1.10.0',
    'google-cloud-datastore>=1.8.0',
    'google-cloud-bigquery>=1.16.0',
    'google-cloud-storage>=1.15.0',
    'google-cloud-kms>=1.2.1',
    'gcsfs>=0.1.2',
    'pyarrow>=0.12.1 ',
    'pandas>=0.24.1',
    'redis>=3.0.1',
    'google-auth>=1.5.0',
    'pyjwt>=1.7.1',
    'cryptography>=2.8',
    'google-cloud-secret-manager>=2.0.0',
    'google-cloud-pubsub>=2.1.0',
    'kfp>=1.6.4'
]

setuptools.setup(
    name="mms-pip",
    version="0.8.4.5",
    author="Josef Goppold, Tobias Hoke",
    author_email="goppold@mediamarktsaturn.com, hoke@mediamarktsaturn.com",
    description="A custom MMS Analytics module for Python3 by the Touchpoint Analytics & Data Discovery",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MediaMarktSaturn/mms-pip",
    packages=setuptools.find_packages(),
    install_requires=required_packages,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)



import subprocess
import google.auth
import google.oauth2.credentials
import google.oauth2.id_token
import google.auth.transport.requests
import time
import jwt
import urllib
import requests
import json

'''
Links:
- https://cloud.google.com/run/docs/authenticating/service-to-service
- https://readthedocs.org/projects/google-auth/downloads/pdf/stable/
- https://medium.com/@stephen.darling/oauth2-authentication-with-google-cloud-run-700015a092c2
- https://medium.com/google-cloud/authenticating-using-google-openid-connect-tokens-e7675051213b
'''


class GCPBearerToken(object):
    def __init__(self, authentication_method, target_url=None, path_to_service_account_file=None, service_account_dictionary=None):
        # 1. "gcloud-sdk": from local machine via gcloud sdk
        # 2. "service-account-file": from a local service-account json file
        # 3. "service-account-dictionary": from a parsed service-account json file
        # 4. "cloudrun-to-cloudrun": https://cloud.google.com/run/docs/authenticating/service-to-service
        # 5. "compute-default":  App Engine, Cloud Run, ‘Compute Engine‘_, or has application default credentials set via GOOGLE_APPLICATION_CREDENTIALS environment variable

        self.__token = None
        self.__exp = None
        self.authentication_method = authentication_method
        self.target_url = target_url
        self.path_to_service_account_file = path_to_service_account_file
        self.service_account_dictionary = service_account_dictionary

        allowed_methods = ["gcloud-sdk", "service-account-file", "service-account-dictionary", "cloudrun-to-cloudrun", "compute-default"]

        # Validate authentication_method:
        if authentication_method not in allowed_methods:
            raise ValueError("authentication_method must be one of {}".format(allowed_methods))

        # Validate required fields:
        if authentication_method != "gcloud-sdk" and target_url is None:
            raise ValueError("Please specify 'target_url' for authentication_method={}".format(authentication_method))

        if authentication_method == "service-account-file" and path_to_service_account_file is None:
            raise ValueError("authentication_method='service-account-file' requires the field 'path_to_service_account_file'")

        if authentication_method == "service-account-dictionary" and service_account_dictionary is None:
            raise ValueError("authentication_method='service-account-dictionary' requires the field 'service_account_dictionary'")

    # ****** gloucd sdk ************************************************************************************************
    @staticmethod
    def __create_from_gcloud_sdk():
        bash_result = subprocess.run(['gcloud', 'auth', 'print-identity-token'], stdout=subprocess.PIPE)
        token = bash_result.stdout
        return token.decode("utf-8")
    # ******************************************************************************************************************

    # ****** service accounts ******************************************************************************************
    @staticmethod
    def __read_sa_file(sa_key_path):
        with open(sa_key_path) as json_file:
            data = json.load(json_file)
        return data

    @staticmethod
    def __create_signed_jwt(credentials, service_url):
        iat = time.time()
        exp = iat + 3600
        payload = {
            'iss': credentials['client_email'],
            'sub': credentials['client_email'],
            'target_audience': service_url,
            'aud': 'https://www.googleapis.com/oauth2/v4/token',
            'iat': iat,
            'exp': exp
        }
        additional_headers = {'kid': credentials['private_key_id']}
        return jwt.encode(payload, credentials['private_key'], headers=additional_headers, algorithm='RS256')

    @staticmethod
    def __exchange_jwt_for_token(signed_jwt):
        body = {
            'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer',
            'assertion': signed_jwt
        }
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        token_request = requests.post(
            url='https://www.googleapis.com/oauth2/v4/token',
            headers=headers,
            data=urllib.parse.urlencode(body)
        )
        result = token_request.json()
        return result['id_token']

    def __create_from_sa(self, credentials):
        # Create signed jwt:
        signed_jwt = self.__create_signed_jwt(credentials, self.target_url)
        # Get id token:
        return self.__exchange_jwt_for_token(signed_jwt)
    # ******************************************************************************************************************

    # ****** cloudrun to cloudrun **************************************************************************************
    def __create_from_cr_to_cr(self):
        # Set up metadata server request
        # See https://cloud.google.com/compute/docs/instances/verifying-instance-identity#request_signature
        metadata_server_token_url = 'http://metadata/computeMetadata/v1/instance/service-accounts/default/identity?audience='
        token_request_url = metadata_server_token_url + self.target_url
        token_request_headers = {'Metadata-Flavor': 'Google'}
        # Fetch the token
        token_response = requests.get(token_request_url, headers=token_request_headers)
        return token_response.content.decode("utf-8")
    # ******************************************************************************************************************

    # ****** cloudrun to cloudrun **************************************************************************************
    def __create_from_compute_default(self):
        request = google.auth.transport.requests.Request()
        return google.oauth2.id_token.fetch_id_token(request, self.target_url)
    # ******************************************************************************************************************

    @staticmethod
    def __get_token_exp(id_token):
        url = "https://oauth2.googleapis.com/tokeninfo?id_token={}".format(id_token)
        r = requests.get(url=url)
        token_info = json.loads(r.content.decode('utf-8'))
        return float(token_info.get("exp"))

    def get_token(self):
        time_now = time.time()
        # Check if token needs initial creation or an update:
        if (self.__token is None) or (self.__exp is None) or (time_now >= self.__exp):
            if self.authentication_method == "gcloud-sdk":
                self.__token = self.__create_from_gcloud_sdk()
            elif self.authentication_method == "service-account-file":
                credentials = self.__read_sa_file(self.path_to_service_account_file)
                self.__token = self.__create_from_sa(credentials)
            elif self.authentication_method == "service-account-dictionary":
                self.__token = self.__create_from_sa(self.service_account_dictionary)
            elif self.authentication_method == "cloudrun-to-cloudrun":
                self.__token = self.__create_from_cr_to_cr()
            elif self.authentication_method == "compute-default":
                self.__token = self.__create_from_compute_default()
            else:
                pass
            self.__exp = self.__get_token_exp(self.__token)
        return self.__token

